#ifndef _DISP2D_H
#define _DISP2D_H
// iVRAMjɐ肷NX
// RXgN^widthQTUȊOƂ܂Ȃ\
//
#include <nds.h>
#include <math.h>
#include <string.h>
class disp2D {
protected :
	const static int OUTNUM = 0x7fffffff ;
	const static double PI = 3.1415926535 ;
	u16 m_offset ;
	u16 *m_vram  ;
	u16 *m_back ;
	int m_width ;
	int m_hight ;
	int m_maxsize ;
public :
	disp2D(int width=256,int hight=192) {
		m_offset= 0 ;
		m_back = NULL ;
		m_width = width ;
		m_hight = hight ;
		m_maxsize = m_width*m_hight ;
	}
	virtual ~disp2D() {
	}
	void setSize(int width,int hight) {
		m_width = width ;
		m_hight = hight ;
		m_maxsize = m_width*m_hight ;
	}
	void setOffset(int ypos) {
		m_offset = ypos ;
	}
	void set2D(u16 *vram) {
		m_vram = vram ;
	}
	u16 *get2D() {
		return m_vram ;
	}
	u16 *get2DBack() {
		return m_back ;
	}
	
	// dot write z֐
	virtual void dot(int x,int y, u16 c) = 0 ;
	virtual void dmaLine(int x,int y,int width,u16 *c) = 0 ;
	virtual void makeBack() = 0 ;
	virtual void getBack() = 0 ;
	virtual void flushBack() = 0 ;
	
	inline static double radian(double s) {
		return PI*(s)/180 ;
	}
	inline static void turn(int s,int x0,int y0,int *x,int *y) {
		double r ;
		int orgx,orgy ;
		orgx = *x ;
		orgy = *y ;
		r = radian((double)s) ;
		*x = (orgx-x0)*cos(r) - (orgy-y0)*sin(r)+x0 ;
		*y = (orgx-x0)*sin(r) + (orgy-y0)*cos(r)+y0 ;
	}
	
//	inline void dot(int x,int y, u16 c) {
//		m_vram[(y<<8)+x] = c ;
//	}
	//   x,y point 
	//   v 
	//   *00
	//   000
	//   000
	inline void dot(int x,int y, u16 c,int width) {
		for( int i = 0 ; i < width ; i++ ) {
			for( int j = 0 ; j < width ; j++ ) {
				dot(x+i,y+j,c) ;
			}
		}
	}
	// levelLine write ( tox(toy)<0  ->  fromx=tox(fromy=toy) )
	inline void levelLine(int fromx,int fromy,int tox,int toy, u16 c,int width=1) {
		int add ;
		if( tox == OUTNUM ) {
			add = (fromy<toy)?1:-1;
			for( int w = 0 ; w < width ; w++ ) {
				for( int y = fromy ; y != toy ; y+=add ) dot(fromx+w,y,c) ;
				dot(fromx+w,toy,c) ;
			}
		}
		if( toy == OUTNUM ) {
			add = (fromx<tox)?1:-1;
			for( int w = 0 ; w < width ; w++ ) {
				for( int x = fromx ; x != tox ; x+=add ) dot(x,fromy+w,c) ;
				dot(tox,fromy+w,c) ;
			}
		}
	}
	inline void line(int fromx,int fromy,int tox,int toy, u16 c,int width=1) {
		int addx,addy ;
		int vx,vy ;
		int x,y ;
		float wf ;
		float af ;
		if( fromx<tox ) {
			addx = 1 ;
			vx = tox-fromx ;
		}
		else {
			addx = -1 ;
			vx = fromx-tox ;
		}
		if( fromy<toy ) {
			addy = 1 ;
			vy = toy-fromy ;
		}
		else {
			addy = -1 ;
			vy = fromy-toy ;
		}
		x = fromx ;
		y = fromy ;
		wf = 0 ;
		if( vx < vy ) {
			af = (float)vx/(float)vy;
			for( int i = 0 ; i < vy ; i++ ) {
				y+=addy ;
				wf+=af ;
				if( wf > 1 ) {
					x+=addx ;
					wf-=1.0 ;
				}
				for( int w = 0 ; w < width ; w++ ) {
					dot(x+w,y,c) ;
				}
			}
		}
		else {
			af = (float)vy/(float)vx;
			for( int i = 0 ; i < vx ; i++ ) {
				x+=addx ;
				wf+=af ;
				if( wf > 1 ) {
					y+=addy ;
					wf-=1.0 ;
				}
				for( int w = 0 ; w < width ; w++ ) {
					dot(x,y+w,c) ;
				}
			}
		}
	}
	inline void fill(int left,int top,int right,int bottom,u16 c) {
		for( int x = left ; x < right ; x++ ) {
			for( int y = top ; y < bottom ; y++ ) {
				dot(x,y,c) ;
			}
		}
	}
	inline void fill(int left,int top,int right,int bottom,u16 *lineData) {
		int width = right-left ;
		for( int y = top ; y < bottom ; y++ ) {
			dmaLine(left,y,width,lineData) ;
		}
	}
	inline void box(int left,int top,int right,int bottom,u16 c,int width=1) {
		levelLine(left,top,right,OUTNUM,c,width) ;
		levelLine(left,top,OUTNUM,bottom,c,width) ;
		levelLine(right,bottom-(width-1),left,OUTNUM,c,width) ;
		levelLine(right-(width-1),bottom,OUTNUM,top,c,width) ;
	}
	inline void circle( int x0, int y0, int r, u16 c,int width=1)
	{
		int x = r;
		int y = 0;
		int F = -2 * r + 3;

		while ( x >= y ) {
			for( int w =0 ; w < width ; w++ ) {
				dot( x0 + x +w, y0 + y, c );
				dot( x0 - x +w, y0 + y, c );
				dot( x0 + x +w, y0 - y, c );
				dot( x0 - x +w, y0 - y, c );
				dot( x0 + y +w, y0 + x, c );
				dot( x0 - y +w, y0 + x, c );
				dot( x0 + y +w, y0 - x, c );
				dot( x0 - y +w, y0 - x, c );
			}
		if ( F >= 0 ) {
			x--;
			F -= 4 * x;
		}
		y++;
		F += 4 * y + 2;
		}
	}

} ;
class disp2D_BMP16 :public disp2D {
protected :
//	const static int max = 192*256 ;
public :
	disp2D_BMP16(int width=256,int hight=192) {
		m_offset = 0 ;
		m_back = NULL ;
		m_width = width ;
		m_hight = hight ;
		m_maxsize = m_width*m_hight ;
	}
	virtual ~disp2D_BMP16() {
		if( m_back != NULL ) delete[] m_back ;
	}
	inline bool check(int *x, int *y) {
		*y += m_offset ;
		if( *x < 0 ) return false ;
		if( (m_width-1) < *x ) return false ;
		if( *y < 0 ) return false ;
		if( (m_hight-1) < *y ) return false ;
		return true ;
	}
	inline void autoSet(int *x,int *y ) {
		*y += m_offset ;
		if( *x < 0 ) *x = 0 ;
		if( (m_width-1) < *x ) *x = (m_width-1) ;
		if( *y < 0 ) *y = 0 ;
		if( (m_hight-1) < *y ) *y = (m_hight-1) ;
	}
	inline void dot(int x,int y, u16 c) {
		if( ! check(&x,&y) ) return ;
		if( m_back == NULL ) m_vram[(y<<8)+x] = c ;
		else m_back[(y<<8)+x] = c ;
	}
	inline void dmaLine(int x,int y,int width,u16 *c) {
		autoSet(&x,&y) ;
		if( m_back == NULL ) dmaCopy(c,&m_vram[(y<<8)+x],width*2) ;
		else dmaCopy(c,&m_back[(y<<8)+x],width*2) ;
	}
	inline void makeBack() {
		m_back = new u16[m_maxsize] ;
//		printf("m_back [%0x]\n",m_back) ;
	}
	inline void getBack() {
		if( m_back == NULL ) makeBack() ;
		dmaCopy(m_vram,m_back,m_maxsize*2) ;
	}
	inline void flushBack() {
		if( m_back == NULL ) makeBack() ;
		dmaCopy(m_back,m_vram,m_maxsize*2) ;
	}
	inline void flushMem(u16 *mem) {
		if( m_back == NULL ) dmaCopy(mem,m_vram,m_maxsize*2) ;
		else dmaCopy(mem,m_back,m_maxsize*2) ;
	}
	inline void flushMem(u16 *mem,int ypos,int ylines) {
		if( m_back == NULL ) dmaCopy(mem,&m_vram[ypos*m_width*2],m_width*ylines*2) ;
		else dmaCopy(mem,&m_back[ypos*m_width*2],m_width*ylines*2) ;
	}
	
	inline void alphaBack() {
		if( m_back == NULL ) return ;
		u16 c ;
		u16 a,r,g,b ;
		for( int i = 0 ; i < m_maxsize ; i++ ) {
			c = m_back[i] ;
			a = (c&0x8000) ;
			r = ((c&0x001F) >>1) & 0x001f;
			g = ((c&0x03e0) >>1) & 0x03e0 ;
			b = ((c&0x7c00) >>1) & 0x7c00 ;
			c = a|r|g|b; 
			m_back[i] = c ;
		}
	}
	inline void clearBack(u16 c) {
		for( int i = 0 ; i < m_maxsize ; i++ ) {
			m_back[i] = c;
		}
	}
} ;
/*
class disp2D_BMP8 :public disp2D {
public :
	disp2D_BMP8() {
	}
	virtual ~disp2D_BMP8() {
	}
	inline void dot(int x,int y, u16 c) {
		u8 c8 ;
		c8 = c&0x00ff ;
		u16 *p = &m_vram[(y<<7)+(x>>1)] ;
		*p = (x&1)?((c8<<8)|(*p&0x00ff)):(c8|(*p&0xff00)) ;
	}
	inline void dmaLine(int x,int y,int width,u16 *c) {
//		dmaCopy(c,&m_vram[(y<<8)+x],width*2) ;
	}
} ;
*/
#endif
